from typing import Optional, List, Union
from pydantic import BaseModel, Field
from threading import Thread, Lock
import json
from langchain_core.language_models import BaseChatModel

from deepeval.utils import trimToJson
from deepeval.test_case import LLMTestCase
from deepeval.metrics import BaseMetric
from deepeval.models import GPTModel
from deepeval.templates import ContextualRelevancyTemplate
from deepeval.progress_context import metrics_progress_context


class ContextualRelevancyVerdict(BaseModel):
    verdict: str
    sentence: str = Field(default=None)


class ContextualRelevancyMetric(BaseMetric):
    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, BaseChatModel]] = None,
        include_reason: bool = True,
    ):
        self.threshold = threshold
        self.model = GPTModel(model=model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason

    def measure(self, test_case: LLMTestCase) -> float:
        if (
            test_case.input is None
            or test_case.actual_output is None
            or test_case.retrieval_context is None
        ):
            raise ValueError(
                "Input, actual output, or retrieval context cannot be None"
            )
        with metrics_progress_context(self.__name__, self.evaluation_model):
            self.verdicts_list: List[
                List[ContextualRelevancyVerdict]
            ] = self._generate_verdicts_list(
                test_case.input, test_case.retrieval_context
            )
            contextual_recall_score = self._generate_score()

            self.reason = self._generate_reason(
                test_case.input, contextual_recall_score
            )

            self.success = contextual_recall_score >= self.threshold
            self.score = contextual_recall_score

            return self.score

    def _generate_reason(self, input: str, score: float):
        if self.include_reason is False:
            return None

        irrelevant_sentences = []
        for index, verdicts in enumerate(self.verdicts_list):
            for verdict in verdicts:
                if verdict.verdict.strip().lower() == "no":
                    data = {"Node": index + 1, "Sentence": verdict.sentence}
                    irrelevant_sentences.append(data)

        prompt: dict = ContextualRelevancyTemplate.generate_reason(
            input=input,
            irrelevant_sentences=irrelevant_sentences,
            score=format(score, ".2f"),
        )

        res = self.model(prompt)
        return res.content

    def _generate_score(self):
        irrelevant_sentences = 0
        total_sentence_count = 0
        for verdicts in self.verdicts_list:
            for verdict in verdicts:
                total_sentence_count += 1
                if verdict.verdict.lower() == "no":
                    irrelevant_sentences += 1

        if total_sentence_count == 0:
            return 0

        return (
            total_sentence_count - irrelevant_sentences
        ) / total_sentence_count

    def _generate_verdicts(
        self,
        text: str,
        context: str,
        verdicts_list: List[List[ContextualRelevancyVerdict]],
        lock: Lock,
    ):
        prompt = ContextualRelevancyTemplate.generate_verdicts(
            text=text, context=context
        )

        res = self.model(prompt)
        json_output = trimToJson(res.content)
        data = json.loads(json_output)
        verdicts = [
            ContextualRelevancyVerdict(**item) for item in data["verdicts"]
        ]

        with lock:
            verdicts_list.append(verdicts)

    def _generate_verdicts_list(
        self, text: str, retrieval_context: List[str]
    ) -> List[List[ContextualRelevancyVerdict]]:
        verdicts_list: List[List[ContextualRelevancyVerdict]] = []
        threads = []
        lock = Lock()
        for context in retrieval_context:
            thread = Thread(
                target=self._generate_verdicts,
                args=(text, context, verdicts_list, lock),
            )
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        return verdicts_list

    def is_successful(self) -> bool:
        self.success = self.score >= self.threshold
        return self.success

    @property
    def __name__(self):
        return "Contextual Relevancy"



"""Tests for toxic
"""


# import pytest
# from deepeval.test_case import LLMTestCase, LLMTestCaseParams
# from deepeval.metrics import ToxicityMetric
# from deepeval import assert_test

# output = "Winners of the FIFA world cup were the French national football team"

# TODO: uncomment once detoxify fixed
# def test_non_toxic():
#     metric = ToxicityMetric(evaluation_params=[LLMTestCaseParams.INPUT])
#     test_case = LLMTestCase(
#         input="Who won the FIFA World Cup in 2018?", actual_output="Don't know."
#     )
#     assert_test(test_case, [metric])


# def test_non_toxic_2():
#     test_case = LLMTestCase(input="Placeholder", actual_output="You're stupid")
#     metric = ToxicityMetric(evaluation_params=[LLMTestCaseParams.ACTUAL_OUTPUT])
#     with pytest.raises(AssertionError):
#         assert_test(test_case, [metric])


# def test_non_toxic_metric():
#     metric = ToxicityMetric(evaluation_params=[LLMTestCaseParams.ACTUAL_OUTPUT])
#     test_case = LLMTestCase(input="placeholder", actual_output=output)
#     assert_test(test_case, [metric])



from pydantic import BaseModel, Field
from typing import Optional, List


class MetricsMetadata(BaseModel):
    metric: str
    score: float
    threshold: float
    success: bool
    reason: Optional[str] = None
    evaluation_model: Optional[str] = Field(None, alias="evaluationModel")


class APITestCase(BaseModel):
    name: str
    input: str
    actual_output: str = Field(..., alias="actualOutput")
    expected_output: Optional[str] = Field(None, alias="expectedOutput")
    success: bool
    metrics_metadata: List[MetricsMetadata] = Field(
        ..., alias="metricsMetadata"
    )
    run_duration: float = Field(..., alias="runDuration")
    traceStack: Optional[dict] = Field(None)
    context: Optional[list] = Field(None)
    retrieval_context: Optional[list] = Field(None, alias="retrievalContext")
    id: Optional[str] = None


class TestRunHttpResponse(BaseModel):
    testRunId: str
    projectId: str
    link: str



from .base_metric import BaseMetric

from .answer_relevancy import AnswerRelevancyMetric
from .base_metric import BaseMetric
from .summarization import SummarizationMetric
from .judgemental_gpt import JudgementalGPT
from .g_eval import GEval
from .faithfulness import FaithfulnessMetric
from .contextual_recall import ContextualRecallMetric
from .contextual_relevancy import ContextualRelevancyMetric
from .contextual_precision import ContextualPrecisionMetric
from .latency import LatencyMetric
from .cost import CostMetric

# from .ragas_metric import (
#     RagasMetric,
#     RAGASAnswerRelevancyMetric,
#     RAGASFaithfulnessMetric,
#     RAGASContextualRecallMetric,
#     RAGASContextualRelevancyMetric,
#     RAGASContextualPrecisionMetric,
#     RAGASAnswerRelevancyMetric,
#     RAGASConcisenessMetric as ConcisenessMetric,
#     RAGASCorrectnessMetric as CorrectnessMetric,
#     RAGASCoherenceMetric as CoherenceMetric,
#     RAGASMaliciousnessMetric as MaliciousnessMetric,
# )
from .bias import BiasMetric
from .toxicity import ToxicityMetric
from .hallucination_metric import HallucinationMetric



from deepeval.models.base import DeepEvalBaseModel
from deepeval.models.summac_model import SummaCModels

# TODO: uncomment out once fixed
# from deepeval.models.detoxify_model import DetoxifyModel
# from deepeval.models.unbias_model import UnBiasedModel

# TODO: restructure or delete (if model logic not needed)
# from deepeval.models.answer_relevancy_model import (
#     AnswerRelevancyModel,
#     CrossEncoderAnswerRelevancyModel,
# )
from deepeval.models.gpt_model import GPTModel



from deepeval.metrics import LatencyMetric
from deepeval.test_case import LLMTestCase
from deepeval import assert_test


def test_latency_metric():
    metric = LatencyMetric(threshold=12)
    test_case = LLMTestCase(
        input="...",
        actual_output="...",
        latency=8.3,
    )
    assert_test(test_case, [metric])



from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase


class LatencyMetric(BaseMetric):
    def __init__(self, threshold: float):
        self.threshold = threshold

    def measure(self, test_case: LLMTestCase):
        self.success = test_case.latency <= self.threshold
        self.score = test_case.latency
        return self.score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Latency"



from deepeval.metrics import CostMetric
from deepeval.test_case import LLMTestCase
from deepeval import assert_test


def test_cost_metric():
    metric = CostMetric(threshold=12)
    test_case = LLMTestCase(input="...", actual_output="...", cost=12)
    assert_test(test_case, [metric])



import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics import HallucinationMetric
from deepeval import assert_test


def test_hallucination_metric():
    metric = HallucinationMetric(threshold=0.5)
    test_case = LLMTestCase(
        input="placeholder",
        actual_output="A blond drinking water in public.",
        context=[
            "A man with blond-hair, and a brown shirt drinking out of a public water fountain."
        ],
    )
    assert_test(test_case, [metric])


def test_hallucination_metric_2():
    metric = HallucinationMetric(threshold=0.6)
    test_case = LLMTestCase(
        input="placeholder",
        actual_output="Python is a programming language.",
        context=["Python is NOT a programming language."],
    )
    with pytest.raises(AssertionError):
        assert_test(test_case, [metric])


def test_hallucination_metric_3():
    metric = HallucinationMetric(threshold=0.6)
    test_case = LLMTestCase(
        input="placeholder",
        actual_output="Python is a programming language.",
        context=["Python is a snake."],
    )
    with pytest.raises(AssertionError):
        assert_test(test_case, [metric])



from typing import Optional, List, Union
from pydantic import BaseModel, Field
import json
from langchain_core.language_models import BaseChatModel

from deepeval.utils import trimToJson
from deepeval.test_case import LLMTestCase
from deepeval.metrics import BaseMetric
from deepeval.models import GPTModel
from deepeval.templates import ContextualRecallTemplate
from deepeval.progress_context import metrics_progress_context


class ContextualRecallVerdict(BaseModel):
    verdict: str
    reason: str = Field(default=None)


class ContextualRecallMetric(BaseMetric):
    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, BaseChatModel]] = None,
        include_reason: bool = True,
    ):
        self.threshold = threshold
        self.model = GPTModel(model=model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason
        self.n = 5

    def measure(self, test_case: LLMTestCase) -> float:
        if (
            test_case.input is None
            or test_case.actual_output is None
            or test_case.retrieval_context is None
            or test_case.expected_output is None
        ):
            raise ValueError(
                "Input, actual output, expected output, or retrieval context cannot be None"
            )
        with metrics_progress_context(self.__name__, self.evaluation_model):
            self.verdicts: List[
                ContextualRecallVerdict
            ] = self._generate_verdicts(
                test_case.expected_output, test_case.retrieval_context
            )

            contextual_recall_score = self._generate_score()

            self.reason = self._generate_reason(
                test_case.expected_output, contextual_recall_score
            )

            self.success = contextual_recall_score >= self.threshold
            self.score = contextual_recall_score
            return self.score

    def _generate_reason(self, expected_output: str, score: float):
        if self.include_reason is False:
            return None

        supportive_reasons = []
        unsupportive_reasons = []
        for verdict in self.verdicts:
            if verdict.verdict.lower() == "yes":
                supportive_reasons.append(verdict.reason)
            else:
                unsupportive_reasons.append(verdict.reason)

        prompt = ContextualRecallTemplate.generate_reason(
            expected_output=expected_output,
            supportive_reasons=supportive_reasons,
            unsupportive_reasons=unsupportive_reasons,
            score=format(score, ".2f"),
        )

        res = self.model(prompt)
        return res.content

    def _generate_score(self):
        if len(self.verdicts) == 0:
            return 0

        justified_sentences = 0
        for verdict in self.verdicts:
            if verdict.verdict.lower() == "yes":
                justified_sentences += 1

        return justified_sentences / len(self.verdicts)

    def _generate_verdicts(
        self, expected_output: str, retrieval_context: List[str]
    ) -> List[ContextualRecallVerdict]:
        prompt = ContextualRecallTemplate.generate_verdicts(
            expected_output=expected_output, retrieval_context=retrieval_context
        )
        res = self.model(prompt)
        json_output = trimToJson(res.content)
        data = json.loads(json_output)
        verdicts = [
            ContextualRecallVerdict(**item) for item in data["verdicts"]
        ]

        return verdicts

    def is_successful(self) -> bool:
        self.success = self.score >= self.threshold
        return self.success

    @property
    def __name__(self):
        return "Contextual Recall"
